/* 
   Unix SMB/CIFS implementation.
   SMB backend for the Common UNIX Printing System ("CUPS")
   Copyright 1999 by Easy Software Products
   Copyright Andrew Tridgell 1994-1998
   Copyright Andrew Bartlett 2002
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2001, 2006 NoMachine, http://www.nomachine.com.          */
/*                                                                        */
/* NXSPOOL, NX protocol compression and NX extensions to this software    */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for applicability.       */
/*                                                                        */
/* NX and NoMachine are trademarks of Medialogic S.p.A.                   */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#define NO_SYSLOG
#undef NX_DEBUG

#include "includes.h"

/*
 * Globals...
 */

extern BOOL		in_client;	/* Boolean for client library */


/*
 * Local functions...
 */

static void		list_devices(void);
static struct cli_state	*smb_connect(const char *, const char *, const char *, const char *, const char *, const char *, BOOL *sessionReq);
static int		smb_print(struct cli_state *, char *, FILE *);

#ifdef NX_DEBUG
FILE *nxdebug;
#endif

/*
 * 'main()' - Main entry for SMB backend.
 */

 int				/* O - Exit status */
 main(int  argc,			/* I - Number of command-line arguments */
     char *argv[])		/* I - Command-line arguments */
{
  int		i;		/* Looping var */
  int		copies;		/* Number of copies */
  char		uri[1024],	/* URI */
		*sep,		/* Pointer to separator */
		*password;	/* Password */
  const char	*username,	/* Username */
		*server,	/* Server name */
		*printer;	/* Printer name */
  const char	*workgroup = "NX";	/* Workgroup */
  const char	*port;		/* Port */
  const char *protocol;
  FILE		*fp;		/* File to print */
  int		status=0;	/* Status of LPD job */
  struct cli_state *cli;	/* SMB interface */
  BOOL sessionReq = True;
  #ifdef NX_DEBUG
  int x;
  #endif


  #ifdef NX_DEBUG
  nxdebug = fopen("/tmp/nxspool.txt", "a+");
  fprintf(nxdebug, "\nNew printing job.\n");
  fflush(nxdebug);
  #endif

  #ifdef NX_DEBUG
  for(x = 0; x <= argc; x++)
  {
     fprintf(nxdebug, "main() argv[%i]: %s\n",x , argv[x]);
  }
  #endif

  /* we expect the URI in argv[0]. Detect the case where it is in argv[1] and cope */
  if (argc > 2 && strncmp(argv[0],"nx://", 5) && !strncmp(argv[1],"nx://", 5)) {
	  argv++;
	  argc--;
  }

  if (argc == 1)
  {
   /*
    * NEW!  In CUPS 1.1 the backends are run with no arguments to list the
    *       available devices.  These can be devices served by this backend
    *       or any other backends (i.e. you can have an SNMP backend that
    *       is only used to enumerate the available network printers... :)
    */

    list_devices();
    return (0);
  }

  if (argc < 6 || argc > 7)
  {
    fprintf(stderr, "Usage: %s [DEVICE_URI] job-id user title copies options [file]\n",
            argv[0]);
    fputs("       The DEVICE_URI environment variable can also contain the\n", stderr);
    fputs("       destination printer:\n", stderr);
    fputs("\n", stderr);
    fputs("           nx://[username:password@][workgroup/]server/printer\n", stderr);
    return (1);
  }

 /*
  * If we have 7 arguments, print the file named on the command-line.
  * Otherwise, print data from stdin...
  */

  if (argc == 6)
  {
   /*
    * Print from Copy stdin to a temporary file...
    */

    fp     = stdin;
    copies = 1;
  }
  else if ((fp = fopen(argv[6], "rb")) == NULL)
  {
    perror("ERROR: Unable to open print file");
    return (1);
  }
  else
    copies = atoi(argv[4]);

  #ifdef NX_DEBUG
  fprintf(nxdebug, "main() copies: %i\n", copies);
  #endif

 /*
  * Find the URI...
  */

  if (getenv("DEVICE_URI") != NULL)
    strncpy(uri, getenv("DEVICE_URI"), sizeof(uri) - 1);
  else if (strncmp(argv[0], "nx://", 5) == 0)
    strncpy(uri, argv[0], sizeof(uri) - 1);
  else
  {
    fputs("ERROR: No device URI found in argv[0] or DEVICE_URI environment variable!\n", stderr);
    return (1);
  }

  uri[sizeof(uri) - 1] = '\0';

  #ifdef NX_DEBUG
  fprintf(nxdebug, "main() uri: %s\n", uri);
  #endif

 /*
  * Extract the destination from the URI...
  */

  if ((sep = strrchr_m(uri, '@')) != NULL)
  {
    username = uri + 5;
    *sep++ = '\0';


    server = sep;

   /*
    * Extract password as needed...
    */

    if ((password = strchr_m(username, ':')) != NULL)
      *password++ = '\0';
    else
      password = "";
  }
  else
  {
    username = "";
    password = "";
    server   = uri + 5;
  }
  #ifdef NX_DEBUG
  fprintf(nxdebug, "main() username: %s\n", username);
  fprintf(nxdebug, "main() password: %s\n", password);
  fprintf(nxdebug, "main() server: %s\n", server);
  #endif

  if ((sep = strchr_m(server, ':')) == NULL)
  {
	  port="0";
  	if ((sep = strchr_m(server, '/')) == NULL)
  	{
    		fputs("ERROR: Bad URI - need printer name!\n", stderr);
    		return (1);
  	}
  }
  else
  {
  	*sep++ = '\0';
  	port = sep;

  	if ((sep = strchr_m(port, '/')) == NULL)
  	{
    		fputs("ERROR: Bad URI - need printer name!\n", stderr);
    		return (1);
  	}
  }

  *sep++ = '\0';
  printer = sep;

  if ((sep = strchr_m(printer, '/')) != NULL)
  {
    *sep++ = '\0';

    protocol  = printer;
    printer   = sep;
  }
  else
    protocol = "cifs";

  #ifdef NX_DEBUG
  fprintf(nxdebug, "main() protocol: %s\n",protocol);
  fprintf(nxdebug, "main() server: %s\n",server);
  fprintf(nxdebug, "main() printer: %s\n",printer);
  fprintf(nxdebug, "main() port: %s\n",port);
  #endif

 /*
  * Setup the SAMBA server state...
  */
/*  setup_logging("smbspool", True);*/

  in_client = True;   /* Make sure that we tell lp_load we are */

/*  if (!lp_load(dyn_CONFIGFILE, True, False, False))
  {
    fprintf(stderr, "ERROR: Can't load %s - run testparm to debug it\n", dyn_CONFIGFILE);
    return (1);
  }
*/

  #ifdef NX_DEBUG
  fprintf(nxdebug, "main() workgroup: %s\n", workgroup);
  #endif

load_interfaces();

  if (protocol == NULL)
  {
    sessionReq = True; 
  }
  else if (strcmp(protocol, "cifs") == 0)
  {
    sessionReq = False;
  }

  do
  {
    if ((cli = smb_connect(workgroup, server, printer, username, password,port, sessionReq)) == NULL)
    {
      if (getenv("CLASS") == NULL)
      {
        fprintf(stderr, "ERROR: Unable to connect to SAMBA host, will retry in 60 seconds...");
        sleep (60);
      }
      else
      {
        fprintf(stderr, "ERROR: Unable to connect to SAMBA host, trying next printer...");
        return (1);
      }
    }
  }
  while (cli == NULL);

 /*
  * Now that we are connected to the server, ignore SIGTERM so that we
  * can finish out any page data the driver sends (e.g. to eject the
  * current page...  Only ignore SIGTERM if we are printing data from
  * stdin (otherwise you can't cancel raw jobs...)
  */

  if (argc < 7)
    CatchSignal(SIGTERM, SIG_IGN);

 /*
  * Queue the job...
  */

  for (i = 0; i < copies; i ++)
    if ((status = smb_print(cli, argv[3] /* title */, fp)) != 0)
      break;

  cli_shutdown(cli);

 /*
  * Return the queue status...
  */

  return (status);
}


/*
 * 'list_devices()' - List the available printers seen on the network...
 */

static void
list_devices(void)
{
 /*
  * Eventually, search the local workgroup for available hosts and printers.
  */

  puts("network nx \"Unknown\" \"Windows Printer via SAMBA/NX\"");
}


/*
 * 'smb_connect()' - Return a connection to a server.
 */

static struct cli_state *		/* O - SMB connection */
smb_connect(const char *workgroup,		/* I - Workgroup */
            const char *server,		/* I - Server */
            const char *share,		/* I - Printer */
            const char *username,		/* I - Username */
            const char *password,		/* I - Password */
            const char *port,		/* I - Port */
            BOOL *sessionReq)
{
  struct cli_state	*c;		/* New connection */
  pstring		myname;		/* Client name */
  NTSTATUS nt_status;
  struct in_addr dest_ip;
  dest_ip=*interpret_addr2("127.0.0.1");

 /*
  * Get the names and addresses of the client and server...
  */

  get_myname(myname);  

  #ifdef NX_DEBUG
  fprintf(nxdebug, "smb_connect() myname: %s\n", myname);
  fprintf(nxdebug, "smb_connect() server: %s\n", server);
  fprintf(nxdebug, "smb_connect() port: %s\n", port);
  fprintf(nxdebug, "smb_connect() share: %s\n", share);
  fprintf(nxdebug, "smb_connect() username: %s\n", username);
  fprintf(nxdebug, "smb_connect() workgroup: %s\n", workgroup);
  fprintf(nxdebug, "smb_connect() password: %s\n",  password);
  fflush(nxdebug);
  #endif

  	
  nt_status = cli_full_connection(&c, myname, server, &dest_ip, atoi(port), share, "?????", 
				  username, workgroup, password, 0, Undefined, NULL, sessionReq);
  
  if (!NT_STATUS_IS_OK(nt_status)) {
	  fprintf(stderr, "ERROR:  Connection failed with error %s\n", nt_errstr(nt_status));
	  return NULL;
  }

  /*
   * Return the new connection...
   */
  
  return (c);
}


/*
 * 'smb_print()' - Queue a job for printing using the SMB protocol.
 */

static int				/* O - 0 = success, non-0 = failure */
smb_print(struct cli_state *cli,	/* I - SMB connection */
          char             *title,	/* I - Title/job name */
          FILE             *fp)		/* I - File to print */
{
  int	fnum;		/* File number */
  int	nbytes,		/* Number of bytes read */
	tbytes;		/* Total bytes read */
  char	buffer[8192],	/* Buffer for copy */
	*ptr;		/* Pointer into tile */


 /*
  * Sanitize the title...
  */

  for (ptr = title; *ptr; ptr ++)
    if (!isalnum((int)*ptr) && !isspace((int)*ptr))
      *ptr = '_';

 /*
  * Open the printer device...
  */

  if ((fnum = cli_open(cli, title, O_RDWR | O_CREAT | O_TRUNC, DENY_NONE)) == -1)
  {
    fprintf(stderr, "ERROR: %s opening remote file %s\n",
            cli_errstr(cli), title);
    return (1);
  }

 /*
  * Copy the file to the printer...
  */

  if (fp != stdin)
    rewind(fp);

  tbytes = 0;

  while ((nbytes = fread(buffer, 1, sizeof(buffer), fp)) > 0)
  {
    if (cli_write(cli, fnum, 0, buffer, tbytes, nbytes) != nbytes)
    {
      fprintf(stderr, "ERROR: Error writing file: %s\n", cli_errstr(cli));
      break;
    }

    tbytes += nbytes;
  } 

  if (!cli_close(cli, fnum))
  {
    fprintf(stderr, "ERROR: %s closing remote file %s\n",
            cli_errstr(cli), title);
    return (1);
  }
  else
    return (0);
}
